/*
	HIMEMC.C
	
	commandline handling for HIMEM
	
	Copyright (c) by tom ehlert 2001-2004 - all rights reserved

	Licensed under the Artistic License version
	
	please see LICENSE.TXT for details

	
*/

#define PROGRAM "HIMEM"

#include "dos.h"
#include "useful.h"


extern char   far init_message[];
extern char   far vinit_message[];
extern char   far copyright[];
extern ushort far xms_num_handles;	/* default 32 */
extern char   far startup_verbose;
extern char   far xms_logging_enabled;
extern char   far no_above_16;

int SoftIce_present(void);


/* called on startup.
	handle commandline "NUMHANDLES=" ...
	...
	mode = 0 if called as driver
	mode = 1 if called as EXE
	
	return: 0         - everything fine
	        errorcode - exit code/abort driver
*/ 

int TheRealMain(int mode, char far *commandline)
{
	char far *found;
	int int15 = 0;
	
	if (mode == EXECMODE_EXE)
		startup_verbose = 1;
    
                             
	printf("%Fs", init_message); 
	printf(" ["__DATE__ "] ");
	printf("%Fs",copyright); 
	printf("\n");

	/******* commandline handling **********/
	
	if (FindCommand(commandline, "/VERBOSE", &found) ||
		StartupLogging())
		{ 
		startup_verbose = 1;
		}              
		
		
		
	if (FindCommand(commandline, "/LOG", &found) )
		{ 
		xms_logging_enabled = 1;
		}

	if (startup_verbose)
		{
		printf("%Fs\n",vinit_message);
		}

	if (FindCommand(commandline, "/NUMHANDLES=", &found) )
		{ 
		xms_num_handles = GetValue(found,10);
		if (startup_verbose)
			printf("selected num_handles=%d\n",xms_num_handles);

		if (xms_num_handles < 8)
			{
			printf("HIMEM:NUMHANDLES must be >= 8, corrected\n");
			xms_num_handles = 8;
			}			
			
		if (xms_num_handles > 128)
			{
			printf("HIMEM:NUMHANDLES must be <= 128, corrected\n");
			xms_num_handles = 128;
			}			
			
		}
		
		
	if (FindCommand(commandline, "/TESTMEM:ON", &found) ||
		FindCommand(commandline, "/TESTMEM:OFF", &found) )
		{
		if (startup_verbose)
			printf("/TESTMEM:ON|OFF not implemented - ignored\n");
		}
		

	if (FindCommand(commandline, "/NOABOVE16", &found) )
		{
			no_above_16 = 1;
		}
	else
		{
		if (SoftIce_present())
			{
			if (startup_verbose)
				printf("SoftIce detected; disabling memory > 64M\n");
			no_above_16 = 1;
			}

		}		
		

	if (FindCommand(commandline, "/INT15=", &found) )
		{ 
		int15 = (int)GetValue(found,16);
		
						/* you probably intended this something to do
						   so we warn (beep) */
		
		printf("\aHIMEM:/INT15=%x - not implemented\n",int15);
		}

	if (FindCommand(commandline, "/TEST", &found))
		{
		extern XMSTESTmain(void);
		
		XMSTESTmain();
		
		return 0;
		}


	if (FindCommand(commandline, "/KILL64", &found))
		{
		extern XMSKILL64(void);
		
		XMSKILL64();
		
		return 0;
		}


	/******* commandline handling done, are there remainders **********/

	commandline = skipWhite(commandline);

	if (*commandline)
		printf("ignored commandline <%Fs>\n", commandline);



	if (mode == EXECMODE_EXE)
		{
							/* called as exe from commandline 
								do :
									usage info
									status report
									show memory statistic
									...
							*/
							
		printf("\n" 
				"HIMEM.EXE is a devicedriver and must be loaded in CONFIG.SYS\n"
				"please load HIMEM as DEVICE=HIMEM.EXE in config.sys\n"
				"\n");

		printf("commandline options available for driver\n"
		       "   /NUMHANDLES=32  default = 32, min = 8, max=128)\n"
		       "   /VERBOSE        more (debug) ouput during installation\n"
		       "   /NOABOVE16      use only low 64M\n"
			   "for executable\n"	
		       "   /TEST           tests all memory + XMS implementation\n"
				);				
				
		return 1;
		}											



	return 0;			/* driver can't fail */

}


#ifdef __TURBOC__
    #pragma warn -asc           /* suppress warning 'switching to assembly */
    #define _asm asm
    void __int__(int);
#endif    

int SoftIce_present(void)
{
#if 1
	int retval = 0;
	
	_asm int 3;

	_asm push si;
	_asm push di;
	
	_asm mov  ah,0x09;
	_asm mov  si,0x4647;	/* ('F'<< 8) | 'G'; */
	_asm mov  di,0x4a4d;	/* ('J'<< 8) | 'M'; */

	_asm mov al,0x12;		/* get breakpoint information */
	
	_asm mov dx,0xffff;
	_asm mov bx,0xffff;
	
	_asm int 3;
	
	_asm pop di;
	_asm pop si;
	
	
	_asm cmp bx,0xffff;
	_asm je not_loaded;
	_asm cmp dx,0xffff;
	_asm je not_loaded;
	
	_asm mov ax,1;	
	_asm mov retval,ax;

not_loaded:	     

	return retval;	


#else
	union REGS ir;

	ir.h.ah = 0x09;		/* S-ICE */

	ir.x.si = ('F'<< 8) | 'G';
	ir.x.di = ('J'<< 8) | 'M';

	ir.h.al = 0x12;		/* get breakpoint information */
	
	ir.x.dx = 0xffff;
	ir.x.bx = 0xffff;
	int86(0x03,&ir,&ir);

	return ir.x.dx != 0xffff && ir.x.bx != 0xffff;
#endif	

}
 
 
 