

/* #define _IODEBUG 1 */
/* #define _IORDEBUG 1 */


/* io.c  -  Virtual disk input/output */

/* Written 1993 by Werner Almesberger */

/*
 * Thu Feb 26 01:15:36 CET 1998: Martin Schulze <joey@infodrom.north.de>
 *	Fixed nasty bug that caused every file with a name like
 *	xxxxxxxx.xxx to be treated as bad name that needed to be fixed.
 */

/* FAT32, VFAT, Atari format support, and various fixes additions May 1998
 * by Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de> */

/* connection to the sector-based DOS disk driver by Eric Auer 2002 */
/* for the beginning, only READING is supported and only FAT12/16   */

/* April 2003 changed by Imre Leber to support reading and writing to disk in FreeDOS */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <fcntl.h>
#ifndef __DJGPP__
#include <linux/fd.h>
#else
/* *** #include "inc/fd.h" seems not to be used anyway *** */
#endif


#include "dosfsck.h"
#include "common.h"
#include "io.h"
#include "volume.h"


typedef struct _change {
    void *data;
    loff_t pos;
    int size;
    struct _change *next;
} CHANGE;

static CHANGE *changes,*last;
static int fd,did_change = 0;

unsigned device_no;

/* ********************************************************** */

static loff_t llseek( int fd, loff_t offset, int whence )
{
    return VolumeSeek(offset);
}

void fs_open(char *path,int rw)
{
    struct stat stbuf;
    
    if ((fd = OpenVolume(path)) < 0)
	pdie("open %s",path);

    changes = last = NULL;
    did_change = 0;

    if (IsWorkingOnImageFile()) {
        if (fstat(GetVolumeHandle(),&stbuf) < 0)
	    pdie("fstat",path);
    }

    /* *** used by boot.c: 2 means it is FAT12 if it looks so... *** */
    device_no = 2;
}

/* ********************************************************** */

void fs_read(loff_t pos,size_t size,void *data)
{
    CHANGE *walk;
    ssize_t got;

    if (llseek(fd,pos,0) != pos) pdie("Seek to %lld",pos);
    if ((got = ReadVolume(data,size)) < 0) pdie("Read %d bytes at %lld",size,pos);
    if (got != size) die("Got %d bytes instead of %d at %lld",got,size,pos);

    for (walk = changes; walk; walk = walk->next) {
	if (walk->pos < pos+size && walk->pos+walk->size > pos) {
	    if (walk->pos < pos)
		memcpy(data,(char *) walk->data+pos-walk->pos,min(size,
		  walk->size-pos+walk->pos));
	    else memcpy((char *) data+walk->pos-pos,walk->data,min(walk->size,
		  size+pos-walk->pos));
	}
    }
}

/* ********************************************************** */

int fs_test(loff_t pos,size_t size)
{
    void *scratch;
    int okay;

    if (llseek(fd,pos,0) != pos) pdie("Seek to %lld",pos);
    scratch = alloc(size);
    okay = ReadVolume(scratch,size) == size;

    free(scratch);
    return okay;
}

/* ********************************************************** */

void fs_write(loff_t pos,size_t size,void *data)
{
    CHANGE *new;
    int did;

    if (write_immed) {
	did_change = 1;

	if (llseek(fd,pos,0) != pos) pdie("Seek to %lld",pos);
	if ((did = WriteVolume(data,size)) == size) return;
	if (did < 0) pdie("Write %d bytes at %lld",size,pos);
	die("Wrote %d bytes instead of %d at %lld",did,size,pos);
    }
    
    new = alloc(sizeof(CHANGE));
    new->pos = pos;
    memcpy(new->data = alloc(new->size = size),data,size);
    new->next = NULL;
    if (last) last->next = new;
    else changes = new;
    last = new;
}

/* ********************************************************** */

static void fs_flush(void)
{
    CHANGE *this;
    int size;

    while (changes) {
	this = changes;
	changes = changes->next;

	if (llseek(fd,this->pos,0) != this->pos)
	    fprintf(stderr,"Seek to %lld failed\n  Did not write %d bytes.\n",
		    this->pos,this->size);
	else if ((size = WriteVolume(this->data,this->size)) < 0)
	    fprintf(stderr,"Writing %d bytes at %lld failed\n",this->size,
		    this->pos);
	else if (size != this->size)
	    fprintf(stderr,"Wrote %d bytes instead of %d bytes at %lld."
		    "\n",size,this->size,this->pos);
    }
    free(this->data);
    free(this);
}

/* ********************************************************** */

int fs_close(int write)
{
    CHANGE *next;
    int changed;

    changed = !!changes;
    if (write) fs_flush();
    else while (changes) {
	    next = changes->next;
	    free(changes->data);
	    free(changes);
	    changes = next;
	}

    if (CloseVolume() < 0) pdie("closing file system");
    
    return changed || did_change;
}

/* ********************************************************** */

int fs_changed(void)
{
    return !!changes || did_change;
}

/* Local Variables: */
/* tab-width: 8     */
/* End:             */
