 ; This file is part of LBAcache, the 386/XMS DOS disk cache by
 ; Eric Auer (eric@coli.uni-sb.de), 2001-2003.

 ; LBAcache is free software; you can redistribute it and/or modify
 ; it under the terms of the GNU General Public License as published
 ; by the Free Software Foundation; either version 2 of the License,
 ; or (at your option) any later version.

 ; LBAcache is distributed in the hope that it will be useful,
 ; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ; GNU General Public License for more details.

 ; You should have received a copy of the GNU General Public License
 ; along with LBAcache; if not, write to the Free Software Foundation,
 ; Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ; (or try http://www.gnu.org/licenses/licenses.html at www.gnu.org).

; LBAcache - a hard disk cache based on XMS, 386 only,
; and aware of the 64bit LBA BIOS Int 13 Extensions.
; GPL 2 software by Eric Auer <eric@coli.uni-sb.de> 2001-2003


; include the SYS header, SYS basics and global data

%ifdef COMCACHE

next:	jmp comloader	; 3 bytes
	nop		; 1 byte
	; (above first 4 bytes will be overwritten by magic chain pointer!)

	dw 0		; device type not used
	dw 0		; near strategy call pointer not used
	dw 0		; near interrupt call pointer not used

%else

next	dw -1		; chain pointer to next device driver
	dw -1
attrib	dw 1000000000000100b	; very simple NUL device
stra	dw strat	; strategy  -> prepare, save ES:BX pointer
inr	dw intr		; interrupt -> main procedure

%endif

nam	db 'HDCACHE$'	; our name, WARNING: becomes invalid file
			; name, this is why the "$" is included
			; ("HDCACHE$.*" being invalid is less annoying)

; ---------------------------------------------------------------

; All the stuff up to and including TSRSIZE should not change in
; structure, because external tools rely on this data being there!
; The magic values are the name at offset 0x0a and the xms handle
; right after that being HDCACHE$ and not-zero respectively.
; LBAcache INFO wants havelba and drvselmask to stay at their places, too.

xmshandle	dw 0	; our XMS handle (also for unloading)

pb		dd 0	; buffer for the ES:BX pointer...

xmsvec		dd 0	; call far pointer for XMS services

geometry:
	db 63,255	; for the hard disks 0x80 .. 0x87:
	db 63,255	;   first byte is max sector number,
	db 63,255	;   second byte is max head number
	db 63,255	; *** NEW: 8 drives supported!

	db 63,255
	db 63,255
	db 63,255
	db 63,255
ageometry       dw 0x112        ; *** geometry of A: 
bgeometry       dw 0x112        ; *** geometry of B: 
			; *** remember that UNCACHE has to be adjusted
			; to reflect the different layout! (infolist)
			; now using labels instead of numbers :-)

rdhit	dd 0	; those 4 counters offer some statistics
rdmiss	dd 0
wrhit	dd 0
wrmiss	dd 0

hint	dw table	; pointer to table (hint for debuggers)
tabsz	dw 8	; hint for debuggers: size of table entries
		; high byte is how far sectors must be shifted right
		; to get number of table entries!

sectors	dw 4096		; how many sectors to cache (512 bytes
			; XMS and {4 bytes CHS / 8 bytes LBA version}
			; DOS RAM for each sector) (also for debuggers)
			; NEW binsel.27.11.2001 uses only 8 bytes for
			; each 1<<N sectors, e.g. for 4 sectors :-)

oldvec	dd 0		; the old INT 0x13 VECTOR (also for unloading)


running	dw 0		; 1 means "init already done"
			; 2 means "disabled" (for the dispatcher)

tsrsize	dw eoftsr	; to know what has to stay in RAM

havelba	dw 0		; 1 means LBA BIOS found, 0 means CHS only
			; for rwfunc (actually only for read: When a
			; real disk read is needed, determine whether
			; it should be done with LBA or with CHS).
			; *** NEW semantics: bitfield, like drvselmask

drvselmask	dw -1	; LSB for disk 0x80 etc. - 1 means cached
			; *** NEW: also a bitfield now. Only low byte used.

; ^^^^^^^^ the above part is COMMON to all versions! (very global data!)

; ---------------------------------------------------------------

; vvvvvvvv the below part is only in .SYS versions

%ifndef COMCACHE

	; .8086 (nasm only has BITS 16 and BITS 32 directives)
	; So we must REMEMBER not to touch 386 code until we
	; have identified the host as "PC with a 80386 or better"

strat:	; in theory, we could even copy the data at es:bx here
	mov word [cs:pb+2],es
	mov word [cs:pb],bx
	RETF

intr:	; the main entry point (called after strat by DOS).
	pushf
        push bx
        push es
        les bx,[cs:pb]
        ; *** mov byte [es:bx+0x17],0	; ???
        and word [es:bx+3],0x7d00	; *** not busy / no error
        or word [es:bx+3],100h		; *** ack status
        cmp byte [es:bx+2],0	; is the command INIT ?
	jnz nix
	cmp byte [cs:running],0	; install already done?
	jnz nix			; only allow init once
        jmp install
nix:    pop es
        pop bx
        popf
        RETF

%endif
; ^^^^^^^^ the above part is only in .SYS versions

; ---------------------------------------------------------------

int13new:
	test word [cs:running],1
	jz int13off			; handler disabled?
	jmp NEWI13			; the dispatcher (386 code...)
int13off:
	jmp far [cs:oldvec]		; chain to old handler

	; all stuff up to this line has to stay in RAM in case we were
	; loaded as SYS or in case int 13 was changed in the meantime!

eoftsr:	; just a label to know where the important part ends

; ---------------------------------------------------------------

; vvvvvvvv the below part is only in .COM versions

%ifdef COMCACHE

comloader:
	mov word [0x100],comsignature+0x100-3	; magic chain pointer, offset
	mov [0x102],cs			; magic chain pointer, segment
					; pointer uses PSP as segment!
	jmp comcache			; relative jump!
comsignature:
	db "COMCACHE"	; 3 bytes after where the device link pointer points

nix:	; code used to go TSR, so it must not be overwritten and must be
	; before the discardable / BSS part with the tables. We check the
	; far "EOF" pointer at [es:bx+0x0e], which always is "cs:somewhere"
	; in LBAcache, and go TSR with appropriate size / exit if size is 0.
	; Not bothering to do an extra "going TSR" message anymore.
	; Not checking TSR EOF pointer -segment- anymore.

	mov ax,cs
	sub ax,0x10		; PSP segment is 0x10 before .sys CS.
	push ax			; Target segment is PSP segment.
	mov ax,nixpsp+0x100	; We now go back to .com memory model!
	push ax			; Target offset is corrected nixpsp offset.
	retf			; CS will point to PSP again, jump to nixpsp.

nixpsp:	mov ax,[es:bx+0x0e]	; EOF TSR offset
	mov cl,4
	shr ax,cl		; convert to paragraphs

	mov bx,cs
	mov ds,bx
	mov es,bx

	or ax,ax		; more than 15 bytes TSR?
	jz nixend

nixtsr:	add ax,0x11		; round up (!) and add PSP size
	; (PS: DOS 3.0+ allows minimal 6+ para TSRs with partial PSP)
	mov dx,ax		; TSR size in paragraphs
	mov ax,0x3100		; go resident (TSR), errorlevel 0 for "loaded"
	int 0x21

nixend:	mov ax,0x4c01		; simply exit, errorlevel 1 for "no TSR"
	int 0x21

%endif

; ^^^^^^^^ the above part is only in .COM versions

; ---------------------------------------------------------------
